import numbers

from siquant.dimensions import SIDimensions, dim_str, dim_pow, dim_mul, dim_div

from .quantities import ScalarQuantity

class SIUnit:
    __slots__ = ('_scale', '_dimensions')

    @staticmethod
    def Unit(scale=1.0, kg=0, m=0, s=0, k=0, a=0, mol=0, cd=0):
        return SIUnit(scale, SIDimensions(kg=kg, m=m, s=s, k=k, a=a, mol=mol, cd=cd))

    def __init__(self, scale, dimensions):
        self._scale = scale
        self._dimensions = dimensions

    scale = property(lambda self: self._scale)
    dimensions = property(lambda self: self._dimensions)

    kg = property(lambda self: self._dimensions[0])
    m = property(lambda self: self._dimensions[1])
    s = property(lambda self: self._dimensions[2])
    k = property(lambda self: self._dimensions[3])
    a = property(lambda self: self._dimensions[4])
    mol = property(lambda self: self._dimensions[5])
    cd = property(lambda self: self._dimensions[6])

    def base_units(self):
        return SIUnit(1.0, self._dimensions)

    def compatible(self, units):
        return self._dimensions == units._dimensions

    def __mul__(self, other):
        if isinstance(other, SIUnit):
            return SIUnit(self._scale * other._scale, dim_mul(self._dimensions, other._dimensions))
        if isinstance(other, numbers.Real):
            return ScalarQuantity(other, self)
        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, numbers.Real):
            return ScalarQuantity(other, self)
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, SIUnit):
            return SIUnit(self._scale / other._scale, dim_div(self._dimensions, other._dimensions))
        if isinstance(other, numbers.Real):
            return ScalarQuantity(1 / other, self)
        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, numbers.Real):
            return ScalarQuantity(other, ~self)
        return NotImplemented

    def __pow__(self, exponent):
        if isinstance(exponent, numbers.Real):
            return SIUnit(
                self._scale ** exponent,
                dim_pow(self._dimensions, exponent)
            )
        return NotImplemented

    def __invert__(self):
        return SIUnit(
            1 / self._scale,
            dim_div(SIDimensions(), self._dimensions)
        )

    def __eq__(self, other):
        if isinstance(other, SIUnit):
            return self._scale == other._scale and self._dimensions == other._dimensions
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, SIUnit):
            return self._scale != other._scale or self._dimensions != other._dimensions
        return NotImplemented

    def __str__(self):
        return '%g*%s' % (self._scale, dim_str(self._dimensions))

    def __repr__(self):
        return 'SIUnit(%f, %r)' % (
            self._scale,
            self._dimensions
        )



import unittest

from siquant.units import SIUnit

class UnitTypeTestCase(unittest.TestCase):

    def test_create_meters(self):
        m = SIUnit.Unit(m=1)
        self.assertEqual(m.scale, 1)
        self.assertEqual(m.m, 1)
        self.assertEqual(m.kg, 0)
        self.assertEqual(m.s, 0)
        self.assertEqual(m.mol, 0)
        self.assertEqual(m.a, 0)
        self.assertEqual(m.k, 0)
        self.assertEqual(m.cd, 0)

    def test_create_unitless(self):
        pi = SIUnit.Unit(3.14159)
        self.assertEqual(pi.scale, 3.14159)
        self.assertEqual(pi.m, 0)
        self.assertEqual(pi.kg, 0)
        self.assertEqual(pi.s, 0)
        self.assertEqual(pi.mol, 0)
        self.assertEqual(pi.a, 0)
        self.assertEqual(pi.k, 0)
        self.assertEqual(pi.cd, 0)

    def test_create_kg(self):
        kg = SIUnit.Unit(kg=1)
        self.assertEqual(kg.scale, 1)
        self.assertEqual(kg.m, 0)
        self.assertEqual(kg.kg, 1)
        self.assertEqual(kg.s, 0)
        self.assertEqual(kg.mol, 0)
        self.assertEqual(kg.a, 0)
        self.assertEqual(kg.k, 0)
        self.assertEqual(kg.cd, 0)

    def test_create_s(self):
        s = SIUnit.Unit(s=1)
        self.assertEqual(s.scale, 1)
        self.assertEqual(s.m, 0)
        self.assertEqual(s.kg, 0)
        self.assertEqual(s.s, 1)
        self.assertEqual(s.mol, 0)
        self.assertEqual(s.a, 0)
        self.assertEqual(s.k, 0)
        self.assertEqual(s.cd, 0)

    def test_create_mol(self):
        pass

    def test_create_a(self):
        pass

    def test_create_k(self):
        pass

    def test_create_cd(self):
        pass

    def test_mul_div_unit(self):
        m = SIUnit.Unit(m=1)
        speed = m / SIUnit.Unit(s=1)
        self.assertEqual(speed.m, 1)
        self.assertEqual(speed.s, -1)

        mm = m / SIUnit.Unit(1000)
        self.assertEqual(mm.scale, 1 / 1000)
        self.assertEqual(mm.m, 1)

        self.assertEqual(m, SIUnit.Unit(1000) * mm)

    def test_pow_unit(self):
        pass

    def test_eq_unit(self):
        pass

    def test_compatible_unit(self):
        pass

    def test_base_units(self):
        pass


from siquant.units import SIUnit
from siquant.quantities import Quantity

nano  = SIUnit.Unit(1 / 1e9)
micro = SIUnit.Unit(1 / 1e6)
milli = SIUnit.Unit(1 / 1000)
centi = SIUnit.Unit(1 / 100)
deci  = SIUnit.Unit(1 / 10)
unity = SIUnit.Unit(1)
deca  = SIUnit.Unit(10)
hecta = SIUnit.Unit(100)
kilo  = SIUnit.Unit(1000)
mega  = SIUnit.Unit(1e6)
giga  = SIUnit.Unit(1e9)
tera  = SIUnit.Unit(1e12)

meters = SIUnit.Unit(m=1)
kilograms = SIUnit.Unit(kg=1)
seconds = SIUnit.Unit(s=1)
kelvin = SIUnit.Unit(k=1)
amperes = SIUnit.Unit(a=1)
mols = SIUnit.Unit(mol=1)
candelas = SIUnit.Unit(cd=1)
radians = SIUnit.Unit(1)
steradians = SIUnit.Unit(1)

newtons = kilograms * meters / seconds ** 2
pascals = newtons / meters ** 2
joules = newtons * meters
watts = joules / seconds
coulombs = amperes * seconds
volts = watts / amperes
farads = coulombs / volts
ohms = volts / amperes
siemens = ~ohms
webers = volts * seconds
teslas = webers / meters ** 2
henrys = webers / amperes
lumens = candelas * steradians
lux = lumens / meters ** 2
becquerels = ~seconds

nanometers  = nano * meters
micrometers = micro * meters
millimeters = milli * meters
centimeters = centi * meters
decimeters  = deci * meters
decameters  = deca * meters
hectameters = hecta * meters
kilometers  = kilo * meters

liters = SIUnit.Unit(1000) * centimeters ** 3
milliliters = milli * liters
microliters = micro * liters

grams = kilograms / kilo
milligrams = milli * grams
micrograms = micro * grams
tonnes = SIUnit.Unit(1000) * kilograms
kilotonnes = kilo * tonnes

nanoseconds  = nano * seconds
microseconds = micro * seconds
milliseconds = milli * seconds
minutes = SIUnit.Unit(60) * seconds
hours = SIUnit.Unit(60) * minutes
days = SIUnit.Unit(24) * hours
weeks = SIUnit.Unit(7) * days
years = SIUnit.Unit(365) * days

gals = centimeters / seconds ** 2
g_0 = SIUnit.Unit(9.80665) * meters / seconds ** 2

micropascals = micro * pascals
millipascals = milli * pascals
kilopascals = kilo * pascals
megapascals = mega * pascals
gigapascals = giga * pascals

millijoules = milli * joules
kilojoules = kilo * joules
megajoules = mega * joules
gigajoules = giga * joules
terajoules = tera * joules

milliwatts = milli * watts
kilowatts = kilo * watts
megawatts = mega * watts
gigawatts = giga * watts
terawatts = tera * watts

millivolts = milli * volts
kilovolts = kilo * volts
megavolts = mega * volts
gigavolts = giga * volts
teravolts = tera * volts

micronewtons = micro * newtons
millinewtons = milli * newtons
kilonewtons = kilo * newtons

newton_meters = newtons * meters
kilonewton_meters = kilo * newton_meters

meters_per_second = meters / seconds


# etc ...



import numbers
from functools import total_ordering

class Quantity:
    __slots__ = ('_quantity', '_units')

    quantity = property(lambda self: self._quantity)
    units = property(lambda self: self._units)

    @classmethod
    def As(cls, units):
        def cvt(quantity):
            if isinstance(quantity, cls):
                return quantity.cvt_to(units)
            return cls(quantity, units)
        return cvt

    def __init__(self, quantity, units):
        self._quantity = quantity
        self._units = units

    def get(self):
        return self._quantity

    def get_as(self, units):
        assert (self._units.compatible(units))
        return self._units._scale / units._scale * self._quantity

    def cvt_to(self, units):
        return self.__class__(self.get_as(units), units)

    def normalized(self):
        return self.__class__(self._quantity * self._units._scale, self._units.base_units())

    def __add__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__(self._quantity + other.get_as(self._units), self._units)
        return NotImplemented

    __iadd__ = __add__

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__(self._quantity - other.get_as(self._units), self._units)
        return NotImplemented

    __isub__ = __sub__

    def __neg__(self):
        return self.__class__(-self._quantity, self._units)


    def __bool__(self):
        return bool(self._quantity)

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self._units.compatible(other._units) and self._quantity == other.get_as(self._units)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, type(self)):
            return not self._units.compatible(other._units) or self._quantity != other.get_as(self._units)
        return NotImplemented

    def __str__(self):
        return '%s %s' % (self._quantity, self._units)

    def __repr__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self._quantity, self._units)

@total_ordering
class ScalarQuantity(Quantity):
    __slots__ = ()

    def __abs__(self):
        return ScalarQuantity(abs(self._quantity), self._units)

    def __lt__(self, other):
        if isinstance(other, ScalarQuantity):
            return self._units.compatible(other._units) and self._quantity < other.get_as(self._units)
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, ScalarQuantity):
            return ScalarQuantity(self._quantity * other._quantity, self._units * other._units)
        if isinstance(other, numbers.Real):
            return ScalarQuantity(self._quantity * other, self._units)
        return NotImplemented

    __imul__ = __mul__

    def __rmul__(self, other):
        if isinstance(other, numbers.Real):
            return ScalarQuantity(other * self._quantity, self._units)
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, ScalarQuantity):
            return ScalarQuantity(self._quantity / other._quantity, self._units / other._units)
        if isinstance(other, numbers.Real):
            return ScalarQuantity(self._quantity / other, self._units)
        return NotImplemented

    __itruediv__ = __truediv__

    def __rtruediv__(self, other):
        if isinstance(other, numbers.Real):
            return ScalarQuantity(other / self._quantity, ~self._units)
        return NotImplemented

    def __pow__(self, exponent):
        if isinstance(exponent, numbers.Real):
            return ScalarQuantity(self._quantity ** exponent, self._units ** exponent)
        return NotImplemented

    def __invert__(self):
        return ScalarQuantity(1 / self._quantity, ~self._units)



from siquant.units import SIUnit

inches = SIUnit.Unit(25.4 / 1000, m=1)
thousandths = inches / SIUnit.Unit(1000)
feet = SIUnit.Unit(12) * inches
yards = SIUnit.Unit(3) * feet
chains = SIUnit.Unit(22) * yards
furlongs = SIUnit.Unit(10) * chains
miles = SIUnit.Unit(8) * furlongs

acres = SIUnit.Unit(1 / 640) * miles ** 2
roods = SIUnit.Unit(1 / 4) * acres
perches = SIUnit.Unit(1 / 40) * roods

fluid_ounces = SIUnit.Unit(1.7339) * inches ** 3
gills = SIUnit.Unit(5) * fluid_ounces
pints = SIUnit.Unit(20) * fluid_ounces
quarts = SIUnit.Unit(40) * fluid_ounces
gallons = SIUnit.Unit(160) * fluid_ounces

pounds = SIUnit.Unit(0.45359237, kg=1)
ounces = pounds / SIUnit.Unit(16)
stones = SIUnit.Unit(14) * pounds
tons = SIUnit.Unit(2240) * pounds

#todo: ...


abbreviations = ('kg', 'm', 's', 'k', 'a', 'mol', 'cd')

def SIDimensions(kg=0, m=0, s=0, k=0, a=0, mol=0, cd=0):
    return (kg, m, s, k, a, mol, cd)

def dim_mul(dims1, dims2):
    return (
        dims1[0] + dims2[0],
        dims1[1] + dims2[1],
        dims1[2] + dims2[2],
        dims1[3] + dims2[3],
        dims1[4] + dims2[4],
        dims1[5] + dims2[5],
        dims1[6] + dims2[6]
    )

def dim_div(dims1, dims2):
    return (
        dims1[0] - dims2[0],
        dims1[1] - dims2[1],
        dims1[2] - dims2[2],
        dims1[3] - dims2[3],
        dims1[4] - dims2[4],
        dims1[5] - dims2[5],
        dims1[6] - dims2[6]
    )

def dim_pow(dims, exp):
    return (
        dims[0] * exp,
        dims[1] * exp,
        dims[2] * exp,
        dims[3] * exp,
        dims[4] * exp,
        dims[5] * exp,
        dims[6] * exp
    )

def dim_str(dims):
    return '*'.join(
        '%s**%g' % (unit_abbreviation, power)
        for unit_abbreviation, power in zip(abbreviations, dims)
        if power != 0
    )




import unittest

import numpy as np

from siquant.units import SIUnit
from siquant.quantities import ScalarQuantity
from siquant.systems import si


class ScalarQuantityTestCase(unittest.TestCase):

    def test_create_and_extract(self):
        mass = ScalarQuantity(100, SIUnit.Unit(kg=1))
        acceleration = ScalarQuantity(1, SIUnit.Unit(m=1, s=-2))

        force = 100 * si.newtons

        self.assertEqual(force.get(), 100)
        self.assertEqual(force.get_as(si.kilonewtons), 0.1)
        self.assertEqual(force, mass * acceleration)

    def test_q_pow(self):
        pass

    def test_q_mul(self):
        pass

    def test_q_div(self):
        pass

    def test_q_add(self):
        pass

    def test_q_sub(self):
        pass


import unittest

import siquant.dimensions as d

class DimensionTestCase(unittest.TestCase):

    def test_dim_mul(self):
        m = d.SIDimensions(kg=1)
        a = d.SIDimensions(m=1, s=-2)

        f = d.dim_mul(m, a)
        self.assertEqual(f, d.SIDimensions(kg=1, m=1, s=-2))

        dist = d.SIDimensions(m=1)
        torque = d.dim_mul(f, dist)
        self.assertEqual(torque, d.SIDimensions(kg=1, m=2, s=-2))

    def test_dim_div(self):

        a = d.SIDimensions(m=1, s=-2)
        f = d.SIDimensions(kg=1, m=1, s=-2)

        m = d.dim_div(f, a)
        self.assertEqual(m, d.SIDimensions(kg=1))

        self.assertEqual(d.dim_div(m, m), d.SIDimensions())

    def test_dim_pow(self):

        dist = d.SIDimensions(m=1)
        area = d.dim_pow(dist, 2)
        self.assertEqual(area, d.SIDimensions(m=2))

        sqrt = d.dim_pow(area, 0.5)
        self.assertEqual(sqrt, d.SIDimensions(m=1))


        v = d.dim_pow(dist, 3)
        self.assertEqual(v, d.SIDimensions(m=3))


if __name__ == '__main__':
    unittest.main()
